# See LICENSE.incore for details
import os
import git
import sys
import re
import pytz
from datetime import datetime
from riscof.utils import yaml
import riscof.utils as utils
import collections
import logging
from git import InvalidGitRepositoryError
import riscof.constants as constants
import riscof.arch_test as arch_test

logger = logging.getLogger(__name__)

class DbgenError(Exception):
    pass

isa_regex = re.compile('''RVTEST_ISA\(\"(?P<isa>.*)\"\)''')
case_regex = re.compile('''RVTEST_CASE\((?P<id>.*),\"(?P<cond>.*)\",(?P<cov_label>.*)\)''')

def dirwalk(dir,ignore_dirs=[]):
    '''
        Recursively searches a directory and returns a list of
        relative paths(from the directory) of the files which end with ".S"(excluding any folder named wip).

        :params: dir - The directory in which the files have to be searched for.

        :return: a list of all .S file paths relative to the riscof home.
    '''
    list = []
    for root, dirs, files in os.walk(dir):
        if not any([i in root for i in ignore_dirs]):
            path = root[root.find(dir):] + "/"
            for file in files:
                if file.endswith(".S"):
                    list.append(os.path.join(path, file))
    return list


def orderdict(foo):
    '''
        Creates and returns a sorted dictionary from a given dictionary.
        :params foo: The dictionary which needs to be sorted.

        :type foo: dict

        :return: A dictionary with alphabetically sorted(based on keys) entries.

    '''
    ret = collections.OrderedDict()
    for key in sorted(foo.keys()):
        ret[key] = foo[key]
    return ret


def createdict(file):
    '''
        Parse the file and create the dictionary node for it.

        :param file: The relative path of the test from riscof home.

        :type file: str

        :return: A dictionary containing all the necesary nodes for the file.

        :raise DbgenError: Raised if the test file doesnt adhere to the rules.
    '''
    with open(file, "r") as k:
        lines = k.read().splitlines()
        code_start = False
        part_start = False
        isa = None
        part_number = 'START'
        i = 0
        part_dict = {}
        line_prev = ''
        while i < len(lines):
            line = lines[i]
            i += 1
            line = line.strip()
            if line == "":
                continue
            if (line.startswith("#") or line.startswith("//")):
                continue
            re_search = isa_regex.search(line)
            if re_search is not None:
                isa = re_search.group('isa')
            if "RVTEST_CASE(" in line:
                temp = ''
                lno = i
                if line.endswith('\\'):
                    temp += (line.strip()).replace("\\", '')
                    while (line.endswith('\\')):
                        line = lines[i]
                        i += 1
                        temp += (line.strip()).replace("\\", '')
                else:
                    temp = line.strip()
                re_search = case_regex.search(temp)
                if re_search is not None:
                    part_number = re_search.group("id")
                    conditions = (re_search.group("cond").replace("//",'')).split(";")
                    labels = [l.strip() for l in (re_search.group("cov_label")).split(",")]

                    if (part_number in part_dict.keys()):
                        logger.warning("{}:{}: Incorrect Naming of Test Case after ({})".
                              format(file, lno, part_number))
                        logger.warning("Skipping file {}. This test will not be included in the\
 database.".format(file))
                        raise DbgenError

                    check = []
                    define = []
                    for cond in conditions:
                        cond = cond.strip()
                        if (cond.startswith('check')):
                            check.append(cond)
                        elif (cond.startswith('def')):
                            define.append(cond)
                    part_dict[part_number] = {'check': check, 'define': define,'coverage_labels':labels}
                else:
                    logger.warning("{}:{}: Incorrect Case String ({})".format(file, lno, part_number))
                    logger.warning("Skipping file {}. This test will not be included in the\
 database.".format(file))
                    raise DbgenError
    if (isa is None):
        logger.warning("{}:ISA not specified.".format( file))
        raise DbgenError
    if len(part_dict.keys()) == 0:
        logger.warning("{}: Atleast one part must exist in the test.".format(file))
        raise DbgenError
    return {'isa': str(isa), 'parts': orderdict(part_dict)}

def check_commit(repo, fpath, old_commit):
    commit = next(
        repo.iter_commits(paths=fpath, max_count=1))
    if (str(commit) != old_commit):
        update = True
    return str(commit), update

def generate():
    flist = dirwalk(constants.suite)

    _ , is_repo = arch_test.get_version(constants.suite)

    dbfile = constants.framework_db
    logger.debug("Using "+dbfile)
    if is_repo:
        repo = git.Repo(constants.suite,search_parent_directories=True)
        tree = repo.tree()
    try:
        db = utils.load_yaml(dbfile)
        delkeys = []
        for key in db.keys():
            if key not in list:
                delkeys.append(key)
        for key in delkeys:
            del db[key]
    except FileNotFoundError:
        db = {}
    existing = db.keys()
    new = [x for x in flist if x not in existing]
    deleted_files = [x for x in existing if x not in flist]
    for entry in deleted_files:
        del db[entry]
    if is_repo:
        for fpath in existing:
            commit_id, update = check_commit(repo,fpath,db[fpath]['commit_id'])
            if(update):
                try:
                    temp = createdict(fpath)
                except DbgenError:
                    del db[fpath]
                    continue
                db[fpath] = {'commit_id': commit_id, **temp}
        for fpath in new:
            commit_id, update = check_commit(repo,fpath,'-')
            try:
                temp = createdict(fpath)
            except DbgenError:
                continue
            db[fpath] = {'commit_id': commit_id, **temp}
    else:
        for fpath in new:
            try:
                temp = createdict(fpath)
            except:
                continue
            db[fpath] = {'commit_id': '-', **temp}
    with open(dbfile, "w") as wrfile:
        wrfile.write('# database generated on ' + (datetime.now(pytz.timezone('GMT'))).strftime("%Y-%m-%d %H:%M GMT")+'\n')
        yaml.dump(orderdict(db),
                  wrfile)



import os
import re
import shutil
import subprocess
import shlex
import logging
import random
import string
from string import Template
import sys

import riscof.utils as utils
import riscof.constants as constants
from riscof.pluginTemplate import pluginTemplate

logger = logging.getLogger()

class dutname(pluginTemplate):
    __model__ = "dutname"

    #TODO: please update the below to indicate family, version, etc of your DUT.
    __version__ = "XXX"

    def __init__(self, *args, **kwargs):
        sclass = super().__init__(*args, **kwargs)

        config = kwargs.get('config')

        # If the config node for this DUT is missing or empty. Raise an error. At minimum we need
        # the paths to the ispec and pspec files
        if config is None:
            print("Please enter input file paths in configuration.")
            raise SystemExit

        # In case of an RTL based DUT, this would be point to the final binary executable of your
        # test-bench produced by a simulator (like verilator, vcs, incisive, etc). In case of an iss or
        # emulator, this variable could point to where the iss binary is located. If 'PATH variable
        # is missing in the config.ini we can hardcode the alternate here.
        self.dut_exe = os.path.join(config['PATH'] if 'PATH' in config else "","dutname")

        # Number of parallel jobs that can be spawned off by RISCOF
        # for various actions performed in later functions, specifically to run the tests in
        # parallel on the DUT executable. Can also be used in the build function if required.
        self.num_jobs = str(config['jobs'] if 'jobs' in config else 1)

        # Path to the directory where this python file is located. Collect it from the config.ini
        self.pluginpath=os.path.abspath(config['pluginpath'])

        # Collect the paths to the  riscv-config absed ISA and platform yaml files. One can choose
        # to hardcode these here itself instead of picking it from the config.ini file.
        self.isa_spec = os.path.abspath(config['ispec'])
        self.platform_spec = os.path.abspath(config['pspec'])

        #We capture if the user would like the run the tests on the target or
        #not. If you are interested in just compiling the tests and not running
        #them on the target, then following variable should be set to False
        if 'target_run' in config and config['target_run']=='0':
            self.target_run = False
        else:
            self.target_run = True

        # Return the parameters set above back to RISCOF for further processing.
        return sclass

    def initialise(self, suite, work_dir, archtest_env):

       # capture the working directory. Any artifacts that the DUT creates should be placed in this
       # directory. Other artifacts from the framework and the Reference plugin will also be placed
       # here itself.
       self.work_dir = work_dir

       # capture the architectural test-suite directory.
       self.suite_dir = suite

       # Note the march is not hardwired here, because it will change for each
       # test. Similarly the output elf name and compile macros will be assigned later in the
       # runTests function
       self.compile_cmd = 'riscv{1}-unknown-elf-gcc -march={0} \
         -static -mcmodel=medany -fvisibility=hidden -nostdlib -nostartfiles -g\
         -T '+self.pluginpath+'/env/link.ld\
         -I '+self.pluginpath+'/env/\
         -I ' + archtest_env + ' {2} -o {3} {4}'

       # add more utility snippets here

    def build(self, isa_yaml, platform_yaml):

      # load the isa yaml as a dictionary in python.
      ispec = utils.load_yaml(isa_yaml)['hart0']

      # capture the XLEN value by picking the max value in 'supported_xlen' field of isa yaml. This
      # will be useful in setting integer value in the compiler string (if not already hardcoded);
      self.xlen = ('64' if 64 in ispec['supported_xlen'] else '32')

      # for dutname start building the '--isa' argument. the self.isa is dutnmae specific and may not be
      # useful for all DUTs
      self.isa = 'rv' + self.xlen
      if "I" in ispec["ISA"]:
          self.isa += 'i'
      if "M" in ispec["ISA"]:
          self.isa += 'm'
      if "C" in ispec["ISA"]:
          self.isa += 'c'

      #TODO: The following assumes you are using the riscv-gcc toolchain. If
      #      not please change appropriately
      self.compile_cmd = self.compile_cmd+' -mabi='+('lp64 ' if 64 in ispec['supported_xlen'] else 'ilp32 ')

    def runTests(self, testList):

      # Delete Makefile if it already exists.
      if os.path.exists(self.work_dir+ "/Makefile." + self.name[:-1]):
            os.remove(self.work_dir+ "/Makefile." + self.name[:-1])
      # create an instance the makeUtil class that we will use to create targets.
      make = utils.makeUtil(makefilePath=os.path.join(self.work_dir, "Makefile." + self.name[:-1]))

      # set the make command that will be used. The num_jobs parameter was set in the __init__
      # function earlier
      make.makeCommand = 'make -j' + self.num_jobs

      # we will iterate over each entry in the testList. Each entry node will be refered to by the
      # variable testname.
      for testname in testList:

          # for each testname we get all its fields (as described by the testList format)
          testentry = testList[testname]

          # we capture the path to the assembly file of this test
          test = testentry['test_path']

          # capture the directory where the artifacts of this test will be dumped/created. RISCOF is
          # going to look into this directory for the signature files
          test_dir = testentry['work_dir']

          # name of the elf file after compilation of the test
          elf = 'my.elf'

          # name of the signature file as per requirement of RISCOF. RISCOF expects the signature to
          # be named as DUT-<dut-name>.signature. The below variable creates an absolute path of
          # signature file.
          sig_file = os.path.join(test_dir, self.name[:-1] + ".signature")

          # for each test there are specific compile macros that need to be enabled. The macros in
          # the testList node only contain the macros/values. For the gcc toolchain we need to
          # prefix with "-D". The following does precisely that.
          compile_macros= ' -D' + " -D".join(testentry['macros'])

          # substitute all variables in the compile command that we created in the initialize
          # function
          cmd = self.compile_cmd.format(testentry['isa'].lower(), self.xlen, test, elf, compile_macros)

	  # if the user wants to disable running the tests and only compile the tests, then
	  # the "else" clause is executed below assigning the sim command to simple no action
	  # echo statement.
          if self.target_run:
            # set up the simulation command. Template is for spike. Please change.
            simcmd = self.dut_exe + ' --isa={0} +signature={1} +signature-granularity=4 {2}'.format(self.isa, sig_file, elf)
          else:
            simcmd = 'echo "NO RUN"'

          # concatenate all commands that need to be executed within a make-target.
          execute = '@cd {0}; {1}; {2};'.format(testentry['work_dir'], cmd, simcmd)

          # create a target. The makeutil will create a target with the name "TARGET<num>" where num
          # starts from 0 and increments automatically for each new target that is added
          make.add_target(execute)

      # if you would like to exit the framework once the makefile generation is complete uncomment the
      # following line. Note this will prevent any signature checking or report generation.
      #raise SystemExit

      # once the make-targets are done and the makefile has been created, run all the targets in
      # parallel using the make command set above.
      make.execute_all(self.work_dir)

      # if target runs are not required then we simply exit as this point after running all
      # the makefile targets.
      if not self.target_run:
          raise SystemExit

#The following is an alternate template that can be used instead of the above.
#The following template only uses shell commands to compile and run the tests.

#    def runTests(self, testList):
#
#      # we will iterate over each entry in the testList. Each entry node will be referred to by the
#      # variable testname.
#      for testname in testList:
#
#          logger.debug('Running Test: {0} on DUT'.format(testname))
#          # for each testname we get all its fields (as described by the testList format)
#          testentry = testList[testname]
#
#          # we capture the path to the assembly file of this test
#          test = testentry['test_path']
#
#          # capture the directory where the artifacts of this test will be dumped/created.
#          test_dir = testentry['work_dir']
#
#          # name of the elf file after compilation of the test
#          elf = 'my.elf'
#
#          # name of the signature file as per requirement of RISCOF. RISCOF expects the signature to
#          # be named as DUT-<dut-name>.signature. The below variable creates an absolute path of
#          # signature file.
#          sig_file = os.path.join(test_dir, self.name[:-1] + ".signature")
#
#          # for each test there are specific compile macros that need to be enabled. The macros in
#          # the testList node only contain the macros/values. For the gcc toolchain we need to
#          # prefix with "-D". The following does precisely that.
#          compile_macros= ' -D' + " -D".join(testentry['macros'])
#
#          # collect the march string required for the compiler
#          marchstr = testentry['isa'].lower()
#
#          # substitute all variables in the compile command that we created in the initialize
#          # function
#          cmd = self.compile_cmd.format(marchstr, self.xlen, test, elf, compile_macros)
#
#          # just a simple logger statement that shows up on the terminal
#          logger.debug('Compiling test: ' + test)
#
#          # the following command spawns a process to run the compile command. Note here, we are
#          # changing the directory for this command to that pointed by test_dir. If you would like
#          # the artifacts to be dumped else where change the test_dir variable to the path of your
#          # choice.
#          utils.shellCommand(cmd).run(cwd=test_dir)
#
#          # for debug purposes if you would like stop the DUT plugin after compilation, you can
#          # comment out the lines below and raise a SystemExit
#
#          if self.target_run:
#            # build the command for running the elf on the DUT. In this case we use spike and indicate
#            # the isa arg that we parsed in the build stage, elf filename and signature filename.
#            # Template is for spike. Please change for your DUT
#            execute = self.dut_exe + ' --isa={0} +signature={1} +signature-granularity=4 {2}'.format(self.isa, sig_file, elf)
#            logger.debug('Executing on Spike ' + execute)
#
#          # launch the execute command. Change the test_dir if required.
#          utils.shellCommand(execute).run(cwd=test_dir)
#
#          # post-processing steps can be added here in the template below
#          #postprocess = 'mv {0} temp.sig'.format(sig_file)'
#          #utils.shellCommand(postprocess).run(cwd=test_dir)
#
#      # if target runs are not required then we simply exit as this point after running all
#      # the makefile targets.
#      if not self.target_run:
#          raise SystemExit



import click

import logging
import importlib
from datetime import datetime
import os
import sys
import pytz
import shutil
import configparser
import distutils.dir_util

from jinja2 import Template

#from riscof.log import *
from riscof.__init__ import __version__
import riscv_config.checker as checker
import riscof.framework.main as framework
import riscof.framework.test as test_routines
import riscof.arch_test as archtest
import riscof.dbgen as dbgen
import riscof.utils as utils
import riscof.constants as constants
from riscof.log import logger
from riscv_config.errors import ValidationError
import riscv_isac.coverage as isac
import riscv_isac
import riscv_config

class Context:
    mkdir = True
    config = None
    dut = None
    base = None
    config_dir = None
    isa_file = None
    platform_file = None

def setup_directories(work_dir,skip_rm=False):
    #Creating work directory
    if not os.path.exists(work_dir):
        logger.debug('Creating new work directory: ' + work_dir)
        os.mkdir(work_dir)
    elif not skip_rm:
        logger.debug('Removing old work directory: ' + work_dir)
        shutil.rmtree(work_dir)
        logger.debug('Creating new work directory: ' + work_dir)
        os.mkdir(work_dir)

def read_config(configfile):
    config = configparser.ConfigParser()
    logger.info("Reading configuration from: "+configfile)
    try:
        config.read(configfile)
    except FileNotFoundError as err:
        logger.error(err)
        raise SystemExit
    return config,os.path.dirname(os.path.abspath(configfile))

def prepare_models(config_dir,config):
    riscof_config = config['RISCOF']
    logger.info("Preparing Models")
    try:
        dut_model = riscof_config['DUTPlugin']
        dut_model_path = utils.absolute_path(config_dir, riscof_config['DUTPluginPath'])

        base_model = riscof_config['ReferencePlugin']
        base_model_path = utils.absolute_path(config_dir, riscof_config['ReferencePluginPath'])
    except KeyError as key:
        logger.error("Error in config file. Possible missing keys.")
        logger.error(key)
        raise SystemExit

    logger.debug("Importing " + dut_model + " plugin from: "+str(dut_model_path))
    sys.path.append(dut_model_path)
    try:
        dut_plugin = importlib.import_module("riscof_" + dut_model)
    except ImportError as msg:
        logger.error("Error while importing "+dut_model+".\n"+str(msg))
        raise SystemExit
    dut_class = getattr(dut_plugin, dut_model)
    if dut_model in config:
        dut = dut_class(name="DUT", config=config[dut_model], config_dir=config_dir)
    else:
        dut = dut_class(name="DUT")

    logger.debug("Importing " + base_model + " plugin from: "+str(base_model_path))
    sys.path.append(base_model_path)
    try:
        base_plugin = importlib.import_module("riscof_" + base_model)
    except ImportError as msg:
        logger.error("Error while importing "+base_model+".\n"+str(msg))
        raise SystemExit
    base_class = getattr(base_plugin, base_model)
    if base_model in config:
        base = base_class(name="Reference", config=config[base_model], config_dir=config_dir)
    else:
        base = base_class(name="Reference")
    return dut,base

def opt_to_name(opt):
    return (opt.replace("--","",1)).replace("-","_")

class CustomOption(click.Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set([opt_to_name(a) for a in kwargs.pop('mutually_exclusive', [])])
        self.requires = [set([opt_to_name(x) for x in a]) for a in kwargs.pop('requires', []) if a]
        help = kwargs.get('help', '')
        if self.mutually_exclusive:
            ex_str = ', '.join(self.mutually_exclusive)
            kwargs['help'] = help + (
                ' NOTE: This argument is mutually exclusive with '
                ' arguments: [' + ex_str + '].'
            )
        if self.requires:
            ex_str = '], ['.join([", ".join(a) for a in self.requires])
            kwargs['help'] = help + (
                ' NOTE: This argument requires one of the following sets of '
                ' arguments: [' + ex_str + '].'
            )
        super(CustomOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise click.UsageError(
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(
                    self.name,
                    ', '.join(self.mutually_exclusive)
                )
            )
        if self.requires:
            if (not any([a.intersection(opts) == a for a in self.requires])) and self.name in opts:
                raise click.UsageError(
                    "Illegal usage: `{}` requires one of the following sets of arguments too "
                    "arguments [{}].".format(
                        self.name,
                        '], ['.join([", ".join(a) for a in self.requires])
                    )
                )

        return super(CustomOption, self).handle_parse_result(
            ctx,
            opts,
            args
        )


@click.group()
@click.version_option(prog_name="RISC-V Architectural Test Framework.",version=__version__)
@click.option('--verbose', '-v', default='info', help='Set verbose level', type=click.Choice(['info','error','debug'],case_sensitive=False))
@click.pass_context
def cli(ctx,verbose):
    logger.level(verbose)
    logger.info('****** RISCOF: RISC-V Architectural Test Framework {0} *******'.format(__version__ ))
    logger.info('using riscv_isac version : ' + str(riscv_isac.__version__))
    logger.info('using riscv_config version : ' + str(riscv_config.__version__))
    ctx.obj = Context()



@cli.command('validateyaml',help= 'Validate the Input YAMLs using riscv-config.')
@click.option('--config',type= click.Path(resolve_path=True,exists=True),
                        help='The Path to the config file. [Default=./config.ini]',
                        metavar= 'PATH',
                        default='./config.ini')
@click.option('--work-dir',default="./riscof_work",metavar='PATH',
        type=click.Path(resolve_path=True,writable=True),
        help='Path to the work directory.')
@click.pass_context
def validate(ctx,config,work_dir):
    if ctx.obj.mkdir:
        setup_directories(work_dir)
        ctx.obj.mkdir = False
    if ctx.obj.config is None:
        ctx.obj.config, ctx.obj.config_dir = read_config(config)
    if ctx.obj.dut is None:
        ctx.obj.dut,ctx.obj.base = prepare_models(ctx.obj.config_dir,ctx.obj.config)
    isa_file = ctx.obj.dut.isa_spec
    platform_file = ctx.obj.dut.platform_spec
    try:
        isa_file = checker.check_isa_specs( isa_file, work_dir, True)
        platform_file = checker.check_platform_specs( platform_file, work_dir, True)
    except ValidationError as msg:
        logger.error(msg)
        raise SystemExit
    ctx.obj.isa_file = isa_file
    ctx.obj.platform_file = platform_file

@cli.command('gendb',help='Generate Database for the Suite.')
@click.option('--suite',required=True,type=click.Path(resolve_path=True,exists=True),metavar="PATH",
    help="Path to the custom Suite Directory.")
@click.option('--env',required=True,type=click.Path(resolve_path=True,exists=True),metavar="PATH",
    help='Path to the env directory for the suite.')
@click.option('--work-dir',default="./riscof_work",metavar='PATH',
    type=click.Path(resolve_path=True,writable=True),
    help='Path to the work directory. [Default = ./riscof_work]')
@click.pass_context
def generate(ctx,suite,env,work_dir):
    if ctx.obj.mkdir:
        setup_directories(work_dir)
        ctx.obj.mkdir = False

    logger.info("Generating database for suite: "+suite)
    constants.suite = suite
    constants.framework_db = os.path.join(work_dir,"database.yaml")
    logger.debug('Suite used: '+constants.suite)
    logger.debug('ENV used: '+ env)
    dbgen.generate()
    logger.info('Database File Generated: '+constants.framework_db)
    constants.env = env
    logger.info('Env path set to'+constants.env)

@cli.command('testlist',help='Generate the test list for the given DUT and suite.')
@click.option('--suite',required=True,type=click.Path(resolve_path=True,exists=True),metavar="PATH",
    help="Path to the custom Suite Directory.")
@click.option('--env',required=True,type=click.Path(resolve_path=True,exists=True),metavar="PATH",
    help='Path to the env directory for the suite.')
@click.option('--config',type= click.Path(resolve_path=True,exists=True),
                        help='The Path to the config file. [Default=./config.ini]',
                        metavar= 'PATH',
                        default='./config.ini')
@click.option('--work-dir',default="./riscof_work",metavar='PATH',
        type=click.Path(resolve_path=True,writable=True),
        help='Path to the work directory. [Default = ./riscof_work]')
@click.pass_context
def testlist(ctx,config,work_dir,suite,env):
    setup_directories(work_dir)
    ctx.obj.mkdir = False
    ctx.invoke(generate,suite=suite,env=env,work_dir=work_dir)
    ctx.obj.gendb = False
    ctx.obj.validate = False
    ctx.invoke(validate,config=config,work_dir=work_dir)
    isa_specs = utils.load_yaml(ctx.obj.isa_file)['hart0']
    platform_specs = utils.load_yaml(ctx.obj.platform_file)
    test_routines.generate_test_pool(isa_specs, platform_specs, work_dir)


@cli.command('run',help='Run the tests on DUT and reference and compare signatures')
@click.option('--suite',required=True,type=click.Path(resolve_path=True,exists=True),metavar="PATH",
    help="Path to the custom Suite Directory.")
@click.option('--env',required=True,type=click.Path(resolve_path=True,exists=True),metavar="PATH",
    help='Path to the env directory for the suite.')
@click.option('--config',type= click.Path(resolve_path=True,exists=True),
                        help='The Path to the config file. [Default=./config.ini]',
                        metavar= 'PATH',
                        default='./config.ini')
@click.option('--work-dir',default="./riscof_work",metavar='PATH',
        type=click.Path(resolve_path=True,writable=True),
        help='Path to the work directory.')
@click.option('--no-browser',is_flag=True,
                     help="Do not open the browser for showing the test report.")
@click.option('--dbfile',type= click.Path(resolve_path=True,exists=True),
                        help='The Path to the database file.',
                        metavar= 'PATH',cls=CustomOption,mutually_exclusive=['--testfile'])
@click.option('--testfile',
                        type= click.Path(resolve_path=True,exists=True),
                        help='The Path to the testlist file.',
                        metavar= 'PATH',cls=CustomOption,mutually_exclusive=['--dbfile'])
@click.option('--no-ref-run',is_flag=True,help="Do not run tests on Reference")
@click.option('--no-dut-run',is_flag=True,help="Do not run tests on DUT")
@click.pass_context
def run(ctx,config,work_dir,suite,env,no_browser,dbfile,testfile,no_ref_run,no_dut_run):
    setup_directories(work_dir,(testfile is not None or dbfile is not None))
    ctx.obj.mkdir = False
    constants.env = env
    constants.suite = suite
    ctx.obj.config, ctx.obj.config_dir = read_config(config)
    ctx.obj.dut,ctx.obj.base = prepare_models(ctx.obj.config_dir,ctx.obj.config)

    dut = ctx.obj.dut
    base = ctx.obj.base

    if testfile is not None or dbfile is not None:
        if dbfile:
            constants.framework_db = dbfile
        else:
            constants.framework_db = os.path.join(work_dir,"database.yaml")
        isa_file = dut.isa_spec
        platform_file = dut.platform_spec
        isa_file = work_dir+ '/' + (isa_file.rsplit('/', 1)[1]).rsplit('.')[0] + "_checked.yaml"
        platform_file = work_dir+ '/' + (platform_file.rsplit('/', 1)[1]).rsplit('.')[0] + "_checked.yaml"
        ctx.obj.isa_file = isa_file
        ctx.obj.platform_file = platform_file
    else:
        ctx.invoke(validate,config=config,work_dir=work_dir)
        ctx.invoke(generate,suite=suite,env=env,work_dir=work_dir)


    with open(ctx.obj.isa_file, "r") as isafile:
        ispecs = isafile.read()
    isa_specs = utils.load_yaml(ctx.obj.isa_file)['hart0']

    with open(ctx.obj.platform_file, "r") as platfile:
        pspecs = platfile.read()

    cntr_args = [dbfile,testfile,no_ref_run,no_dut_run]

    report_objects = {}
    report_objects['date'] = (datetime.now(
        pytz.timezone('GMT'))).strftime("%Y-%m-%d %H:%M GMT")
    report_objects['riscof_version'] = __version__
    report_objects['dut'] = (dut.__model__).replace("_", " ")
    report_objects['reference'] = (base.__model__).replace("_", " ")

    rvarch, _ = archtest.get_version(constants.suite)
    report_objects['rvarch_version'] = rvarch['version'] if rvarch['version'] != "-" else \
                                        rvarch['commit']

    report_objects['isa'] = isa_specs['ISA']
    report_objects['usv'] = isa_specs['User_Spec_Version']
    report_objects['psv'] = isa_specs['Privilege_Spec_Version']
    report_objects['isa_yaml'] = ctx.obj.isa_file
    report_objects['platform_yaml'] = ctx.obj.platform_file
    report_objects['isa_specs'] = ispecs
    report_objects['platform_specs'] = pspecs

    report_objects['results'] = framework.run(dut, base, ctx.obj.isa_file,
                                              ctx.obj.platform_file, work_dir, cntr_args)

    report_objects['num_passed'] = 0
    report_objects['num_failed'] = 0

    for entry in report_objects['results']:
        if entry['res'] == "Passed":
            report_objects['num_passed'] += 1
        else:
            report_objects['num_failed'] += 1

    with open(constants.html_template, "r") as report_template:
        template = Template(report_template.read())

    output = template.render(report_objects)

    reportfile = os.path.join(work_dir, "report.html")
    with open(reportfile, "w") as report:
        report.write(output)

    shutil.copyfile(constants.css,
                    os.path.join(work_dir, "style.css"))

    logger.info("Test report generated at "+reportfile+".")
    if not no_browser:
        try:
            import webbrowser
            logger.info("Openning test report in web-browser")
            webbrowser.open(reportfile)
        except:
            raise SystemExit



@cli.command('coverage',help='Run the tests on DUT and reference and compare signatures')
@click.option('--suite',required=True,type=click.Path(resolve_path=True,exists=True),metavar="PATH",
    help="Path to the custom Suite Directory.")
@click.option('--env',required=True,type=click.Path(resolve_path=True,exists=True),metavar="PATH",
    help='Path to the env directory for the suite.')
@click.option('--config',type= click.Path(resolve_path=True,exists=True),
                        help='The Path to the config file. [Default=./config.ini]',
                        metavar= 'PATH',
                        default='./config.ini')
@click.option('--work-dir',default="./riscof_work",metavar='PATH',
        type=click.Path(resolve_path=True,writable=True),
        help='Path to the work directory. [Default = ./riscof_work]')
@click.option('--no-browser',is_flag=True,
                     help="Do not open the browser for showing the test report.")
@click.option(
        '--cgf-file','-c',multiple=True,
        type=click.Path(resolve_path=True,readable=True,exists=True),
        help="Coverage Group File(s). Multiple allowed.",required=True
    )
@click.pass_context
def coverage(ctx,config,work_dir,suite,env,no_browser,cgf_file):
    setup_directories(work_dir)
    ctx.obj.mkdir = False
    ctx.obj.config, ctx.obj.config_dir = read_config(config)
    ctx.obj.dut,ctx.obj.base = prepare_models(ctx.obj.config_dir,ctx.obj.config)

    dut = ctx.obj.dut
    base = ctx.obj.base
    ctx.invoke(validate,config=config,work_dir=work_dir)
    ctx.invoke(generate,suite=suite,env=env,work_dir=work_dir)
    logger.info('Will collect Coverage using RISCV-ISAC')
    logger.info('CGF file(s) being used : ' + str(cgf_file))
    isa_file = ctx.obj.isa_file
    platform_file = ctx.obj.platform_file

    with open(isa_file, "r") as isafile:
        ispecs = isafile.read()
    isa_specs = utils.load_yaml(ctx.obj.isa_file)['hart0']

    with open(platform_file, "r") as platfile:
        pspecs = platfile.read()
    report, for_html, test_stats, coverpoints = framework.run_coverage(base, isa_file, platform_file,
            work_dir, cgf_file)
    report_file = open(work_dir+'/suite_coverage.rpt','w')
    utils.dump_yaml(report, report_file)
    report_file.close()


    report_objects = {}
    report_objects['date'] = (datetime.now(
        pytz.timezone('GMT'))).strftime("%Y-%m-%d %H:%M GMT")
    report_objects['riscof_version'] = __version__
    report_objects['reference'] = (base.__model__).replace("_", " ")

    rvarch, _ = archtest.get_version(constants.suite)
    report_objects['rvarch_version'] = rvarch['version'] if rvarch['version'] != "-" else \
                                        rvarch['commit']

    report_objects['isa'] = isa_specs['ISA']
    report_objects['usv'] = isa_specs['User_Spec_Version']
    report_objects['psv'] = isa_specs['Privilege_Spec_Version']
    report_objects['isa_yaml'] = isa_file
    report_objects['platform_yaml'] = platform_file
    report_objects['isa_specs'] = ispecs
    report_objects['platform_specs'] = pspecs
    report_objects['results'] = for_html
    report_objects['results1'] = test_stats
    report_objects['coverpoints'] = coverpoints
    with open(constants.coverage_template, "r") as report_template:
        template = Template(report_template.read())

    output = template.render(report_objects)

    reportfile = os.path.join(work_dir, "coverage.html")
    with open(reportfile, "w") as report:
        report.write(output)

    shutil.copyfile(constants.css,
                    os.path.join(work_dir, "style.css"))

    logger.info("Test report generated at "+reportfile+".")
    if not no_browser:
        try:
            import webbrowser
            logger.info("Openning test report in web-browser")
            webbrowser.open(reportfile)
        except:
            raise SystemExit



@cli.command('arch-test',help='Setup and maintenance for Architectural TestSuite.',
    no_args_is_help=True)
@click.option('--dir',type=click.Path(resolve_path=True),
    help="The Path to the directory to initialise/containing the tests. [Default = ./riscv-arch-test]",
    metavar = 'PATH',default="./riscv-arch-test")
@click.option('--get-version',type=str,default='latest',
    help=' Version of the repository to get. [Default = latest]',
    cls=CustomOption,requires=[['--clone'],['--update']])
@click.option('--clone',is_flag=True,
    help="Clone and setup the architectural tests from the remote repository.",cls=CustomOption,
    mutually_exclusive=["--update",'--show-version'])
@click.option('--update',is_flag=True,
    help="Update the architectural tests from the remote repository.",
    cls=CustomOption,mutually_exclusive=["--clone",'--show-version'])
@click.option('--show-version',is_flag=True,
    help="Print version of the local architectural test suite.",
    cls=CustomOption,mutually_exclusive=['--clone','--update'])
def arch_test(dir,get_version,clone,update,show_version):
    if(clone):
        if os.path.exists(dir):
            shutil.rmtree(dir)
        archtest.clone(dir,get_version)
    elif(update):
        archtest.update(dir,get_version)
    elif(show_version):
        version, is_repo = archtest.get_version(dir)
        if not is_repo:
            logger.error("Not the riscv-arch-test repo.")
        else:
            logger.info("Clonned version {0} of the repository with commit hash {1} ".format(
                    version['version'],version['commit']))
    else:
        logger.error("Please specify one of [update,clone,show-version] flags.")



@cli.command('setup',help='Initiate Setup for riscof.')
@click.option('--dutname',type=str,help='Name of DUT plugin. [Default=spike]',
    default='spike',metavar= 'NAME')
@click.option('--refname',type=str,help='Name of Reference plugin. [Default=sail_cSim]',
    default='sail_cSim',metavar= 'NAME')
@click.option('--work-dir',default="./riscof_work",metavar='PATH',
        type=click.Path(resolve_path=True,writable=True),
        help='Path to the work directory.')
def setup(dutname,refname,work_dir):
    logger.info("Setting up sample plugin requirements [Old files will be overwritten]")
    try:
        cwd = os.getcwd()
        logger.info("Creating sample Plugin directory for [DUT]: " +dutname)
        src = os.path.join(constants.root, "Templates/setup/model/")
        dest = os.path.join(cwd, dutname)
        distutils.dir_util.copy_tree(src, dest)

        os.rename(cwd+'/'+dutname+'/model_isa.yaml',
                cwd+'/'+dutname+'/'+dutname+'_isa.yaml')
        os.rename(cwd+'/'+dutname+'/model_platform.yaml',
                cwd+'/'+dutname+'/'+dutname+'_platform.yaml')
        os.rename(cwd+'/'+dutname+'/riscof_model.py',
                cwd+'/'+dutname+'/riscof_'+dutname+'.py')
        with open(cwd+'/'+dutname+'/riscof_'+dutname+'.py', 'r') as file :
          filedata = file.read()

        # Replace the target string
        filedata = filedata.replace('dutname', dutname)

        # Write the file out again
        with open(cwd+'/'+dutname+'/riscof_'+dutname+'.py', 'w') as file:
          file.write(filedata)

        logger.info("Creating sample Plugin directory for [REF]: " +\
              refname)
        if refname == 'sail_cSim':
            src = os.path.join(constants.root, "Templates/setup/sail_cSim/")
            dest = os.path.join(cwd, refname)
            distutils.dir_util.copy_tree(src, dest)
        else:
            src = os.path.join(constants.root, "Templates/setup/reference/")
            dest = os.path.join(cwd, refname)
            distutils.dir_util.copy_tree(src, dest)
            os.rename(cwd+'/'+refname+'/riscof_model.py',
                cwd+'/'+refname+'/riscof_'+refname+'.py')
            with open(cwd+'/'+refname+'/riscof_'+refname+'.py', 'r') as file :
              filedata = file.read()

            # Replace the target string
            filedata = filedata.replace('refname', refname)

            # Write the file out again
            with open(cwd+'/'+refname+'/riscof_'+refname+'.py', 'w') as file:
              file.write(filedata)

        logger.info("Creating Sample Config File")
        configfile = open('config.ini','w')
        configfile.write(constants.config_temp.format(refname, \
                cwd+'/'+refname, dutname,cwd+'/'+dutname))
        logger.info('**NOTE**: Please update the paths of the reference \
and plugins in the config.ini file')
        configfile.close()
    except FileExistsError as err:
        logger.error(err)
        raise SystemExit

if __name__=="__main__":
    cli()



# See LICENSE.incore for details

import logging
import git
from git import InvalidGitRepositoryError
import riscof.constants as constants
logger = logging.getLogger(__name__)

def get_version(path):
    ver_dict = {
        'commit': '-',
        'version': '-'
    }
    success = False
    try:
        repo = git.Repo(path,search_parent_directories=True)
    except InvalidGitRepositoryError:
        logger.debug("Suite path is not a git repository.")
    else:
        commit = repo.head.commit
        tags = repo.tags
        remote = repo.remote()
        url = remote.url
        if (tags) and (url==constants.https_url or url == constants.ssh_url):
            ver_dict['commit'] = str(commit)
            for tag in tags:
                if tag.commit == commit:
                    ver_dict['version'] = str(tag)
            success = True
        else:
            logger.debug("Repository is not the official RISCV Architectural Test suite or the \
branch is not the master branch.")
    return ver_dict,success

def update(path,branch='master'):
    version,is_repo = get_version(path)
    if is_repo:
        logger.debug("Current version of the repository: " + version['version'])
        logger.debug("Current commit hash of the repository: " + version['commit'])
        repo = git.Repo(path)
        repo.git.pull('origin','master')
        latest_tag = (repo.tags)[-1]
        checkout_target = latest_tag if branch=='latest' else branch
        repo.git.checkout(checkout_target)
        version,_ = get_version(path)
        logger.info("Updated version of the repository: " + version['version'])
        logger.info("Updated commit hash of the repository: " + version['commit'])
    else:
        logger.info("Directory does not contain the riscv-arch-test repo.")

def clone(path,branch="master"):
    logger.info("Clonning repository at "+str(path))
    repo = git.Repo.clone_from(constants.https_url, path)
    latest_tag = (repo.tags)[-1]
    checkout_target = latest_tag if branch=='latest' else branch
    repo.git.checkout(checkout_target)
    version, _ = get_version(path)
    logger.info("Clonned version {0} of the repository with commit hash {1} ".format(
                    str(version['version']),version['commit']))




import os

root = os.path.abspath(os.path.dirname(__file__))

suite = os.path.join(root,"suite/")
#cgf_dir = os.path.join(root,"coverage/")
#
#cgf_file = []
#cgf_file.append(os.path.join(root,"coverage/dataset.cgf"))
#for path in os.listdir(cgf_dir):
#    if path != 'dataset.cgf':
#        full_path = os.path.join(cgf_dir, path)
#        if os.path.isfile(full_path):
#            cgf_file.append(full_path)

https_url = 'https://github.com/riscv/riscv-arch-test.git'
ssh_url = 'git@github.com:riscv/riscv-arch-test.git'

framework_db = os.path.join(root, "framework/database.yaml")
cwd = os.getcwd()
#work_dir = os.path.join(cwd, "riscof_work/")
html_template = os.path.join(root, 'Templates/report.html')
coverage_template = os.path.join(root, 'Templates/coverage.html')
css = os.path.join(root, 'Templates/style.css')
env = os.path.join(root, suite + "env/")

config_temp = '''[RISCOF]
ReferencePlugin={0}
ReferencePluginPath={1}
DUTPlugin={2}
DUTPluginPath={3}

[{2}]
pluginpath={3}
ispec={3}/{2}_isa.yaml
pspec={3}/{2}_platform.yaml
target_run=1

[{0}]
pluginpath={1}
'''


